----------------------------------------------------------------------------------------------------

library ieee;
  use ieee.std_logic_1164.all;
  use ieee.std_logic_arith.all;
  use ieee.std_logic_misc.all;
  use ieee.std_logic_unsigned.all;

entity fmc110_cpld is
port (
  fmc_to_cpld     : inout std_logic_vector(3 downto 0); -- SPI communication with host
  spi_sclk        : out   std_logic;                    -- Local SPI clock
  spi_sdio        : inout std_logic;                    -- Local SPI data in/out
  spi_n_cs        : out   std_logic_vector(4 downto 0); -- Local SPI chip selects
  clk_n_reset     : out   std_logic;                    -- SPI reset Clock Tree
  dac_n_reset     : out   std_logic;                    -- SPI reset DAC
  clksrc_sel      : out   std_logic_vector(2 downto 0); -- Clock source selection
  syncsrc_sel     : out   std_logic_vector(1 downto 0); -- Synchronisation source selection
  clk_refmon      : in    std_logic;                    -- Clock Tree reference monitor status
  clk_ld          : in    std_logic;                    -- Clock Tree PLL lock status
  clk_status      : in    std_logic;                    -- Clock Tree programmable status
  vm_n_int        : in    std_logic;                    -- Voltage Monitor alarm interupt
  ps_en           : in    std_logic;                    -- Local Power Supply Enable
  led             : out   std_logic;                    -- Led driver
  eeprom_wp       : out   std_logic;                    -- EEPROM write protect
  front_io_dir    : out   std_logic_vector(3 downto 0); -- Front panel LVTTL IO direction control
  mictor_io       : inout std_logic_vector(3 downto 0); -- Auxilary IO on Mictor connectors
  fan_n_en        : out   std_logic_vector(3 downto 0)  -- Local FAN enable controls
);
end fmc110_cpld;

architecture fmc110_cpld_syn of fmc110_cpld is

-- Address constants:
-- Internal registers (first bit of presel=0)
constant ADDR_REG0 : std_logic_vector(1 downto 0) := "00";
constant ADDR_REG1 : std_logic_vector(1 downto 0) := "01";
constant ADDR_REG2 : std_logic_vector(1 downto 0) := "10";
-- External registers (first bit of presel=1)
constant ADDR_ADC0 : std_logic_vector(2 downto 0) := "000";
constant ADDR_ADC1 : std_logic_vector(2 downto 0) := "001";
constant ADDR_DAC0 : std_logic_vector(2 downto 0) := "010";
constant ADDR_DAC1 : std_logic_vector(2 downto 0) := "011";
constant ADDR_CLK0 : std_logic_vector(2 downto 0) := "100";

signal sclk      : std_logic;
signal n_cs      : std_logic;
signal sdi       : std_logic;
signal sdo       : std_logic;
signal rw        : std_logic;
signal oe        : std_logic;
signal n_irq     : std_logic;

signal cnt       : integer range 0 to 31;
signal shift_in  : std_logic_vector(7 downto 0);
signal shift_out : std_logic_vector(7 downto 0);
signal src_loc   : std_logic; -- location: 0 = internal / 1 = external
signal src_sel   : std_logic_vector(2 downto 0);
signal reg_sel   : std_logic_vector(1 downto 0);

signal reg0_w    : std_logic_vector(7 downto 0) := (others => '0');
signal reg1_w    : std_logic_vector(7 downto 0) := (others => '0');
signal reg2_w    : std_logic_vector(7 downto 0) := (others => '0');

signal reg0_r    : std_logic_vector(7 downto 0) := (others => '0');
signal reg1_r    : std_logic_vector(7 downto 0) := (others => '0');
signal reg2_r    : std_logic_vector(7 downto 0) := (others => '0');

begin

----------------------------------------------------------------------------------------------------
-- Map
----------------------------------------------------------------------------------------------------

sclk <= fmc_to_cpld(0);
n_cs <= fmc_to_cpld(1);
sdi  <= fmc_to_cpld(2);

fmc_to_cpld(2) <= sdo      when (oe = '1' and src_loc = '0') else
                  spi_sdio when (oe = '1' and src_loc = '1') else
                  'Z';

fmc_to_cpld(3) <= n_irq;

----------------------------------------------------------------------------------------------------
-- Shift Source select
----------------------------------------------------------------------------------------------------

process (n_cs, sclk)
begin
  if (n_cs = '1') then

    cnt       <= 0;
    shift_in  <= (others => '0');
    rw        <= '0';

  elsif (rising_edge(sclk)) then

    if (cnt < 31) then
      cnt <= cnt + 1;
    end if;

    shift_in <= shift_in(shift_in'length-2 downto 0) & sdi;

    -- Capter register data on the last rising edge
    if (cnt = 23 and src_loc = '0') then
      case reg_sel is
        when ADDR_REG0 => reg0_w <= shift_in(6 downto 0) & sdi;
        when ADDR_REG1 => reg1_w <= shift_in(6 downto 0) & sdi;
        when ADDR_REG2 => reg2_w <= shift_in(6 downto 0) & sdi;
        when others =>
      end case;
    end if;

    if (cnt = 9) then
      rw <= shift_in(0);
    end if;

  end if;
end process;

process (n_cs, sclk)
begin
  if (n_cs = '1') then

    src_loc   <= '0';
    src_sel   <= (others => '0');
    reg_sel   <= (others => '0');
    oe        <= '0';
    shift_out <= (others => '0');

  elsif (falling_edge(sclk)) then

    if (cnt = 8) then
      src_loc <= shift_in(7);
      src_sel <= shift_in(src_sel'length-1 downto 0);
    end if;

    if (cnt = 16) then
      reg_sel <= shift_in(reg_sel'length-1 downto 0);
    end if;

    if (src_sel = ADDR_CLK0 and cnt = 24) then --Clock tree has 16-bit instruction word
      oe <= transport rw after 1 ps;
    elsif (src_sel /= ADDR_CLK0 and cnt = 16) then  --Others has 8-bit instruction word
      oe <= transport rw after 1 ps;
    end if;

    if (cnt = 16) then
      --case reg_sel is
      case shift_in(reg_sel'length-1 downto 0) is
        when ADDR_REG0 => shift_out <= reg0_r;
        when ADDR_REG1 => shift_out <= reg1_r;
        when ADDR_REG2 => shift_out <= reg2_r;
        when others =>
      end case;
    else
      shift_out <= shift_out(shift_out'length-2 downto 0) & '0';
    end if;

  end if;
end process;

sdo <= shift_out(shift_out'length-1);

----------------------------------------------------------------------------------------------------
-- SPI Bus Enable
----------------------------------------------------------------------------------------------------
spi_sclk <= sclk when src_loc = '1' else '0';
spi_sdio <= sdi  when (oe = '0' and src_loc = '1') else 'Z';

spi_n_cs(0) <= n_cs when (src_loc = '1' and src_sel = ADDR_ADC0) else '1';
spi_n_cs(1) <= n_cs when (src_loc = '1' and src_sel = ADDR_ADC1) else '1';
spi_n_cs(2) <= n_cs when (src_loc = '1' and src_sel = ADDR_DAC0) else '1';
spi_n_cs(3) <= n_cs when (src_loc = '1' and src_sel = ADDR_DAC1) else '1';
spi_n_cs(4) <= n_cs when (src_loc = '1' and src_sel = ADDR_CLK0) else '1';

----------------------------------------------------------------------------------------------------
-- Register controlled outputs
----------------------------------------------------------------------------------------------------
clksrc_sel   <= reg0_w(2 downto 0);
syncsrc_sel  <= reg0_w(4 downto 3);
clk_n_reset  <= not reg0_w(5);
dac_n_reset  <= not reg0_w(6);
eeprom_wp    <= not reg0_w(7);

front_io_dir <= reg1_w(3 downto 0);
fan_n_en     <= reg1_w(7 downto 4);

----------------------------------------------------------------------------------------------------
-- Mictor IO
----------------------------------------------------------------------------------------------------
mictor_io(0) <= 'Z';
mictor_io(1) <= 'Z';
mictor_io(2) <= 'Z';
mictor_io(3) <= 'Z';

----------------------------------------------------------------------------------------------------
-- Interupt
----------------------------------------------------------------------------------------------------
n_irq <= vm_n_int and clk_refmon and clk_ld and clk_status;

----------------------------------------------------------------------------------------------------
-- LED status output
----------------------------------------------------------------------------------------------------
led <= clk_refmon   when reg2_w(0) = '1' else
       clk_ld       when reg2_w(1) = '1' else
       clk_status   when reg2_w(2) = '1' else
       not vm_n_int when reg2_w(3) = '1' else
       not n_irq    when reg2_w(4) = '1' else '0';

----------------------------------------------------------------------------------------------------
-- LED status output
----------------------------------------------------------------------------------------------------
reg0_r <= reg0_w;
reg1_r <= reg1_w;

reg2_r(0) <= clk_refmon;
reg2_r(1) <= clk_ld;
reg2_r(2) <= clk_status;
reg2_r(3) <= not vm_n_int;
reg2_r(4) <= not n_irq;
reg2_r(5) <= '0';
reg2_r(6) <= '0';
reg2_r(7) <= '0';

----------------------------------------------------------------------------------------------------
-- End
----------------------------------------------------------------------------------------------------

end fmc110_cpld_syn;
