-------------------------------------------------------------------------------------
-- FILE NAME : ads5400_phy_sp.vhd
--
-- AUTHOR    : Peter Kortekaas
--
-- COMPANY   : 4DSP
--
-- ITEM      : 1
--
-- UNITS     : Entity       - ads5400_phy_sp
--             architecture - ads5400_phy_sp_syn
--
-- LANGUAGE  : VHDL
--
-------------------------------------------------------------------------------------
--
-------------------------------------------------------------------------------------
-- DESCRIPTION
-- ===========
--
-- ads5400_phy_sp
-- Notes: ads5400_phy_sp
-------------------------------------------------------------------------------------
--  Disclaimer: LIMITED WARRANTY AND DISCLAIMER. These designs are
--              provided to you as is.  4DSP specifically disclaims any
--              implied warranties of merchantability, non-infringement, or
--              fitness for a particular purpose. 4DSP does not warrant that
--              the functions contained in these designs will meet your
--              requirements, or that the operation of these designs will be
--              uninterrupted or error free, or that defects in the Designs
--              will be corrected. Furthermore, 4DSP does not warrant or
--              make any representations regarding use or the results of the
--              use of the designs in terms of correctness, accuracy,
--              reliability, or otherwise.
--
--              LIMITATION OF LIABILITY. In no event will 4DSP or its
--              licensors be liable for any loss of data, lost profits, cost
--              or procurement of substitute goods or services, or for any
--              special, incidental, consequential, or indirect damages
--              arising from the use or operation of the designs or
--              accompanying documentation, however caused and on any theory
--              of liability. This limitation will apply even if 4DSP
--              has been advised of the possibility of such damage. This
--              limitation shall apply not-withstanding the failure of the
--              essential purpose of any limited remedies herein.
--
----------------------------------------------

-- Library declarations
library ieee;
  use ieee.std_logic_1164.all;
  use ieee.std_logic_unsigned.all;
  use ieee.std_logic_arith.all;
  use ieee.std_logic_misc.all;
  use ieee.numeric_std.all;
library unisim;
  use unisim.vcomponents.all;

entity ads5400_phy_sp is
  generic
  (
    START_ADDR   : std_logic_vector(27 downto 0) := x"0000000";
    STOP_ADDR    : std_logic_vector(27 downto 0) := x"00000FF";
    DMUX_MODE    : integer := 1 -- either 1 for 1:1 mode or 2 for 1:2 mode
  );
  port (
    -- Global signals
    rst          : in  std_logic;

    -- Command Interface
    clk_cmd      : in  std_logic;
    in_cmd_val   : in  std_logic;
    in_cmd       : in  std_logic_vector(63 downto 0);
    out_cmd_val  : out std_logic;
    out_cmd      : out std_logic_vector(63 downto 0);
    in_cmd_busy  : out std_logic;

    -- DDR LVDS Interface
    clk_a_p      : in  std_logic;
    clk_a_n      : in  std_logic;

    cha_p        : in  std_logic_vector(11 downto 0);
    cha_n        : in  std_logic_vector(11 downto 0);
    ovra_p       : in  std_logic;
    ovra_n       : in  std_logic;

    -- Output data
    phy_out_clk  : out std_logic;                      -- clock 1/8 of the sample frequecy
    phy_out_data : out std_logic_vector(127 downto 0)  -- 8 samples in parallel in 16-bit format

  );
end ads5400_phy_sp;

architecture ads5400_phy_sp_syn of ads5400_phy_sp is

----------------------------------------------------------------------------------------------------
-- Components
----------------------------------------------------------------------------------------------------
component fmc110_stellar_cmd is
generic
(
  START_ADDR           : std_logic_vector(27 downto 0) := x"0000000";
  STOP_ADDR            : std_logic_vector(27 downto 0) := x"00000FF"
);
port
(
  reset                : in  std_logic;
  -- Command Interface
  clk_cmd              : in  std_logic;                     --cmd_in and cmd_out are synchronous to this clock;
  out_cmd              : out std_logic_vector(63 downto 0);
  out_cmd_val          : out std_logic;
  in_cmd               : in  std_logic_vector(63 downto 0);
  in_cmd_val           : in  std_logic;
  -- Register interface
  clk_reg              : in  std_logic;                     --register interface is synchronous to this clock
  out_reg              : out std_logic_vector(31 downto 0); --caries the out register data
  out_reg_val          : out std_logic;                     --the out_reg has valid data  (pulse)
  out_reg_addr         : out std_logic_vector(27 downto 0); --out register address
  in_reg               : in  std_logic_vector(31 downto 0); --requested register data is placed on this bus
  in_reg_val           : in  std_logic;                     --pulse to indicate requested register is valid
  in_reg_req           : out std_logic;                     --pulse to request data
  in_reg_addr          : out std_logic_vector(27 downto 0);  --requested address
  --mailbox interface
  mbx_in_reg           : in  std_logic_vector(31 downto 0); --value of the mailbox to send
  mbx_in_val           : in  std_logic                      --pulse to indicate mailbox is valid
);
end component fmc110_stellar_cmd;

component serdes is
generic (
  SYS_W                 : integer := 10; -- width of the data for the system
  DEV_W                 : integer := 80  -- width of the data for the device
);
port (
  -- From the system into the device
  data_in_from_pins_p   : in    std_logic_vector(SYS_W-1 downto 0);
  data_in_from_pins_n   : in    std_logic_vector(SYS_W-1 downto 0);
  data_in_to_device     : out   std_logic_vector(DEV_W-1 downto 0);
  -- Input, Output delay control signals
  rst_cmd               : in    std_logic;                            -- command reset
  clk_cmd               : in    std_logic;                            -- command clock
  start_align           : in    std_logic;                            -- start automatic alignment, test pattern should be tuned on
  data_aligned_out      : out   std_logic_vector(SYS_W-1 downto 0);   -- automatic alignment ready flags
  data_aligned          : out   std_logic;                            -- automatic alignment ready flag (and function)
  delay_reset           : in    std_logic;                            -- active high synchronous reset for input delay
  delay_valid_inc       : in    std_logic;                            --
  delay_data_inc        : in    std_logic_vector(SYS_W-1 downto 0);   -- delay increment
  delay_valid_dec       : in    std_logic;                            --
  delay_data_dec        : in    std_logic_vector(SYS_W-1 downto 0);   -- delay decrement
  delay_value           : out   std_logic_vector(5*SYS_W-1 downto 0); -- automatic alignment ready flags
  bitslip               : in    std_logic_vector(SYS_W-1 downto 0);   -- bitslip module is enabled in networking mode, user should tie it to '0' if not needed
  -- Clock and reset signals
  clk_in_p              : in    std_logic;                            -- differential fast clock from iob
  clk_in_n              : in    std_logic;
  clk_div_out           : out   std_logic;                            -- slow clock output
  clk_reset             : in    std_logic;                            -- reset signal for clock circuit
  io_reset              : in    std_logic                             -- reset signal for io circuit
);
end component serdes;

component ads5400_phy_fifo is
port (
  rst           : in std_logic;
  wr_clk        : in std_logic;
  rd_clk        : in std_logic;
  din           : in std_logic_vector(127 downto 0);
  wr_en         : in std_logic;
  rd_en         : in std_logic;
  dout          : out std_logic_vector(127 downto 0);
  full          : out std_logic;
  empty         : out std_logic;
  valid         : out std_logic;
  rd_data_count : out std_logic_vector(3 downto 0);
  wr_data_count : out std_logic_vector(3 downto 0)
);
end component ads5400_phy_fifo;

component pulse2pulse is
port (
  in_clk   : in  std_logic;
  out_clk  : in  std_logic;
  rst      : in  std_logic;
  pulsein  : in  std_logic;
  inbusy   : out std_logic;
  pulseout : out std_logic
);
end component pulse2pulse;

----------------------------------------------------------------------------------------------------
-- Constants
----------------------------------------------------------------------------------------------------
constant ADDR_COMMAND    : std_logic_vector(31 downto 0) := x"00000000";
constant ADDR_CONTROL    : std_logic_vector(31 downto 0) := x"00000001";

constant ADDR_INC_A      : std_logic_vector(31 downto 0) := x"00000002";
constant ADDR_DEC_A      : std_logic_vector(31 downto 0) := x"00000006";
constant ADDR_BITSLIP_A  : std_logic_vector(31 downto 0) := x"0000000A";

constant ADDR_TAP_VAL0   : std_logic_vector(31 downto 0) := x"0000000B";
constant ADDR_TAP_VAL1   : std_logic_vector(31 downto 0) := x"0000000C";
constant ADDR_TAP_VAL2   : std_logic_vector(31 downto 0) := x"0000000D";
constant ADDR_TAP_VAL3   : std_logic_vector(31 downto 0) := x"0000000E";

----------------------------------------------------------------------------------------------------
-- Signals
----------------------------------------------------------------------------------------------------
signal out_reg_val    : std_logic;
signal out_reg_addr   : std_logic_vector(27 downto 0);
signal out_reg        : std_logic_vector(31 downto 0);

signal in_reg_req     : std_logic;
signal in_reg_addr    : std_logic_vector(27 downto 0);
signal in_reg_val     : std_logic;
signal in_reg         : std_logic_vector(31 downto 0);

signal cmd_reg        : std_logic_vector(31 downto 0);

signal sync_cmd       : std_logic;
signal sync_pls       : std_logic;

signal delay_reset    : std_logic;
signal clk_reset      : std_logic;
signal io_reset       : std_logic;

signal start_align    : std_logic;
signal data_aligned   : std_logic;
signal data_aligned_out : std_logic_vector(12 downto 0);

signal in0_p          : std_logic_vector(12 downto 0);
signal in0_n          : std_logic_vector(12 downto 0);
signal out0           : std_logic_vector(103 downto 0);

signal in1_p          : std_logic_vector(9 downto 0);
signal in1_n          : std_logic_vector(9 downto 0);
signal out1           : std_logic_vector(79 downto 0);

signal in2_p          : std_logic_vector(12 downto 0);
signal in2_n          : std_logic_vector(12 downto 0);
signal out2           : std_logic_vector(103 downto 0);

signal in3_p          : std_logic_vector(10 downto 0);
signal in3_n          : std_logic_vector(10 downto 0);
signal out3           : std_logic_vector(87 downto 0);

signal adr            : std_logic;
signal ad             : std_logic_vector(103 downto 0);
signal aor            : std_logic_vector(7 downto 0);

signal bdr            : std_logic;
signal bd             : std_logic_vector(79 downto 0);
signal bor            : std_logic_vector(7 downto 0);

signal cdr            : std_logic;
signal cd             : std_logic_vector(79 downto 0);
signal cor            : std_logic_vector(7 downto 0);

signal ddr            : std_logic;
signal dd             : std_logic_vector(79 downto 0);
signal dor            : std_logic_vector(7 downto 0);

signal ad_8x16        : std_logic_vector(127 downto 0);
signal bd_8x16        : std_logic_vector(127 downto 0);
signal cd_8x16        : std_logic_vector(127 downto 0);
signal dd_8x16        : std_logic_vector(127 downto 0);

signal ad_8x16_o      : std_logic_vector(127 downto 0);
signal bd_8x16_o      : std_logic_vector(127 downto 0);
signal cd_8x16_o      : std_logic_vector(127 downto 0);
signal dd_8x16_o      : std_logic_vector(127 downto 0);

signal delay_inc_a_val: std_logic;
signal delay_inc_a    : std_logic_vector(12 downto 0);
signal delay_dec_a_val: std_logic;
signal delay_dec_a    : std_logic_vector(12 downto 0);
signal delay_value    : std_logic_vector(64 downto 0);
signal bitslip_a      : std_logic_vector(12 downto 0);

signal pec            : std_logic;
signal pef            : std_logic;

signal fifo_empty    :std_logic;
signal wr_en         :std_logic;
signal rd_en         :std_logic;
begin

----------------------------------------------------------------------------------------------------
-- Stellar Command Interface
----------------------------------------------------------------------------------------------------
fmc110_stellar_cmd_inst : fmc110_stellar_cmd
generic map
(
  START_ADDR   => START_ADDR,
  STOP_ADDR    => STOP_ADDR
)
port map
(
  reset        => rst,

  clk_cmd      => clk_cmd,
  in_cmd_val   => in_cmd_val,
  in_cmd       => in_cmd,
  out_cmd_val  => out_cmd_val,
  out_cmd      => out_cmd,

  clk_reg      => clk_cmd,
  out_reg_val  => out_reg_val,
  out_reg_addr => out_reg_addr,
  out_reg      => out_reg,

  in_reg_req   => in_reg_req,
  in_reg_addr  => in_reg_addr,
  in_reg_val   => in_reg_val,
  in_reg       => in_reg,

  mbx_in_val   => '0',
  mbx_in_reg   => (others => '0')
);

in_cmd_busy <= '0';

----------------------------------------------------------------------------------------------------
-- Registers
----------------------------------------------------------------------------------------------------
process (rst, clk_cmd)
begin
  if (rst = '1') then

    cmd_reg        <= (others => '0');
    pec            <= '0';
    delay_inc_a_val<= '0';
    delay_inc_a    <= (others => '0');
    delay_dec_a_val<= '0';
    delay_dec_a    <= (others => '0');
    bitslip_a      <= (others => '0');
    in_reg_val     <= '0';
    in_reg         <= (others=> '0');

  elsif (rising_edge(clk_cmd)) then

    -- Write commands
    if (out_reg_val = '1' and out_reg_addr = ADDR_COMMAND) then
      cmd_reg <= out_reg;
    else
      cmd_reg <= (others => '0');
    end if;

    -- Write controls
    if (out_reg_val = '1' and out_reg_addr = ADDR_CONTROL) then
      --delay_inc_a   <= out_reg(0);
      --force_wr_en   <= out_reg(1);
      --force_reset   <= out_reg(2);
      --force_dcm_rst <= out_reg(3);
      pec <= out_reg(4);
	else
      pec <= '0';
    end if;

    --Increment register
    if (out_reg_val = '1' and out_reg_addr = ADDR_INC_A) then
      delay_inc_a_val <= '1';
      delay_inc_a <= out_reg(delay_inc_a'length-1 downto 0);
    else
      delay_inc_a_val <= '0';
    end if;

    --Decrement register
    if (out_reg_val = '1' and out_reg_addr = ADDR_DEC_A) then
      delay_dec_a_val <= '1';
      delay_dec_a <= out_reg(delay_dec_a'length-1 downto 0);
    else
      delay_dec_a_val <= '0';
    end if;

    if (out_reg_val = '1' and out_reg_addr = ADDR_BITSLIP_A) then
      bitslip_a <= out_reg(bitslip_a'length-1 downto 0);
    end if;

    -- Read
    if (in_reg_req = '1' and in_reg_addr = ADDR_COMMAND) then
      in_reg_val <= '1';
      in_reg(31 downto 16) <= conv_std_logic_vector(0, 16-13) & data_aligned_out;
      in_reg(15 downto  0) <= conv_std_logic_vector(0, 16- 1) & data_aligned;

    elsif (in_reg_req = '1' and in_reg_addr = ADDR_CONTROL) then
      in_reg_val <= '1';
      in_reg     <= conv_std_logic_vector(0, 32-5) & pef & "0000";

    elsif (in_reg_req = '1' and in_reg_addr = ADDR_TAP_VAL0) then
      in_reg_val <= '1';
      in_reg( 7 downto  0) <= conv_std_logic_vector(0, 8-5) & delay_value( 0*5+4 downto  0*5);
      in_reg(15 downto  8) <= conv_std_logic_vector(0, 8-5) & delay_value( 1*5+4 downto  1*5);
      in_reg(23 downto 16) <= conv_std_logic_vector(0, 8-5) & delay_value( 2*5+4 downto  2*5);
      in_reg(31 downto 24) <= conv_std_logic_vector(0, 8-5) & delay_value( 3*5+4 downto  3*5);

    elsif (in_reg_req = '1' and in_reg_addr = ADDR_TAP_VAL1) then
      in_reg_val <= '1';
      in_reg( 7 downto  0) <= conv_std_logic_vector(0, 8-5) & delay_value( 4*5+4 downto  4*5);
      in_reg(15 downto  8) <= conv_std_logic_vector(0, 8-5) & delay_value( 5*5+4 downto  5*5);
      in_reg(23 downto 16) <= conv_std_logic_vector(0, 8-5) & delay_value( 6*5+4 downto  6*5);
      in_reg(31 downto 24) <= conv_std_logic_vector(0, 8-5) & delay_value( 7*5+4 downto  7*5);

	elsif (in_reg_req = '1' and in_reg_addr = ADDR_TAP_VAL2) then
      in_reg_val <= '1';
      in_reg( 7 downto  0) <= conv_std_logic_vector(0, 8-5) & delay_value( 8*5+4 downto  8*5);
      in_reg(15 downto  8) <= conv_std_logic_vector(0, 8-5) & delay_value( 9*5+4 downto  9*5);
      in_reg(23 downto 16) <= conv_std_logic_vector(0, 8-5) & delay_value(10*5+4 downto 10*5);
      in_reg(31 downto 24) <= conv_std_logic_vector(0, 8-5) & delay_value(11*5+4 downto 11*5);

	elsif (in_reg_req = '1' and in_reg_addr = ADDR_TAP_VAL3) then
      in_reg_val <= '1';
      in_reg( 7 downto  0) <= conv_std_logic_vector(0, 8-5) & delay_value(12*5+4 downto 12*5);
      in_reg(15 downto  8) <= conv_std_logic_vector(0, 8-5) & "00000";
      in_reg(23 downto 16) <= conv_std_logic_vector(0, 8-5) & "00000";
      in_reg(31 downto 24) <= conv_std_logic_vector(0, 8-5) & "00000";

    else
      in_reg_val <= '0';
      in_reg     <= in_reg;

    end if;

  end if;
end process;

----------------------------------------------------------------------------------------------------
-- Map commands
----------------------------------------------------------------------------------------------------
delay_reset     <= cmd_reg(0) or rst;
clk_reset       <= cmd_reg(1) or rst;
io_reset        <= cmd_reg(2) or rst;
start_align     <= cmd_reg(3);

----------------------------------------------------------------------------------------------------
-- Serdes input mapping (4 clock domains/regions)
----------------------------------------------------------------------------------------------------
in0_p <= ovra_p & cha_p;
in0_n <= ovra_n & cha_n;

----------------------------------------------------------------------------------------------------
-- Channel A
----------------------------------------------------------------------------------------------------
serdes_inst0 : serdes
generic map (
  SYS_W               => 13,
  DEV_W               => 104
)
port map (
  data_in_from_pins_p => in0_p,
  data_in_from_pins_n => in0_n,
  data_in_to_device   => out0,
  rst_cmd             => rst,
  clk_cmd             => clk_cmd,
  start_align         => start_align,
  data_aligned_out    => data_aligned_out,
  data_aligned        => data_aligned,
  delay_reset         => delay_reset,
  delay_valid_inc     => delay_inc_a_val,
  delay_data_inc      => delay_inc_a,
  delay_valid_dec     => delay_dec_a_val,
  delay_data_dec      => delay_dec_a,
  delay_value         => delay_value,
  bitslip             => bitslip_a,
  clk_in_p            => clk_a_p,
  clk_in_n            => clk_a_n,
  clk_div_out         => adr,
  clk_reset           => clk_reset,
  io_reset            => io_reset
);

----------------------------------------------------------------------------------------------------
-- Serdes output mapping
----------------------------------------------------------------------------------------------------
samples_0: for s in 0 to 7 generate

  data_bits_0: for b in 0 to 11 generate

    ad(s*12+b) <= out0(s*13+b);

  end generate;

  or_bits_0: for b in 0 to 0 generate

    aor(s) <= out0(s*13+12);

  end generate;

end generate;

----------------------------------------------------------------------------------------------------
-- Reorder data
----------------------------------------------------------------------------------------------------
samples: for s in 0 to 7 generate

  zero_bits: for b in 13 to 15 generate

    ad_8x16(s*16+b) <= '0';

  end generate;

  or_bits: for b in 12 to 12 generate

    ad_8x16(s*16+b) <= aor(s);

  end generate;

  data_bits: for b in 0 to 11 generate

    ad_8x16(s*16+b) <= ad(s*12+(b-0));

  end generate;

end generate;

----------------------------------------------------------------------------------------------------
-- FIFO's
----------------------------------------------------------------------------------------------------
ads5400_phy_fifo_a : ads5400_phy_fifo
port map (
  rst           => io_reset,
  wr_clk        => adr,
  rd_clk        => adr,
  din           => ad_8x16,
  wr_en         => '1',
  rd_en         => '1',
  dout          => ad_8x16_o,
  full          => open,
  empty         => fifo_empty,
  valid         => open,
  rd_data_count => open,
  wr_data_count => open
);

----------------------------------------------------------------------------------------------------
-- Pattern check (toggling pattern)
----------------------------------------------------------------------------------------------------
process (pec, adr)
begin
  if (pec = '1') then
    pef <= '0';

  elsif (rising_edge(adr)) then
    if (ad_8x16 /= x"0000_1FFF_0000_1FFF_0000_1FFF_0000_1FFF" and ad_8x16 /= x"1FFF_0000_1FFF_0000_1FFF_0000_1FFF_0000") then
      pef <= '1';
    end if;
  end if;
end process;
process (rst, adr)
begin
  if (rst = '1') then
    wr_en   <= '0';
    rd_en   <= '0';
  elsif (rising_edge(adr)) then

      wr_en   <= '1';
      rd_en  <= not fifo_empty;
  end if;
end process;

----------------------------------------------------------------------------------------------------
-- Connect entity
----------------------------------------------------------------------------------------------------
phy_out_clk  <= adr;
phy_out_data <= ad_8x16_o;

----------------------------------------------------------------------------------------------------
-- end
----------------------------------------------------------------------------------------------------
end ads5400_phy_sp_syn;
